/*
 * Huk: a multi-threaded modular bruteforce tool 
 *
 * Written by Carlos Barros <barros [at] barrossecurity [dot] com>
 *
 * Copyright (C) 2006, Barrossecurity
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>

#include <libhuk/common.h>
#include <libhuk/wrappers.h>
#include <libhuk/libnet.h>
#include <libhuk/list.h>

#include "wlist.h"
#include "dispatcher.h"
#include "brute.h"
#include "session.h"
#include "modules.h"

/*=[ Prototypes ]*/
int brute_newtask(s_dispatcher *);
int brute_task(void *);
s_task *brute_wait_task(s_dispatcher *);
void brute_abort_request(int);
void brute_status_request(int);

/*=[ Global variables ]=*/
extern s_session	*session;		// session object, from main.c

/*
** Public methods
*/

int do_brute_force()
{
	s_dispatcher	*disp;
	s_task		*task;
	int		i;

	// setup signal handlers
	signal(SIGQUIT,brute_abort_request);
	signal(SIGINT,brute_status_request);

	// create new dispatcher
	disp = dispatcher_new(session->max_conn,session->conn_timeout);

	// fill the initial task buffer
	for(i=0;i<session->max_conn;i++)
		if(brute_newtask(disp) == ERROR) break;

	// now, waits one task to exit and dispatch a new one
	if(i == session->max_conn)
	{
		while((task = brute_wait_task(disp)))
		{
			// free slot
			dispatcher_free_slot(disp,task->pid);
	
			// check abort request
			if(session->abort_bruteforce == ABORT_REQUEST)
			{
				session->abort_bruteforce = ABORTED;
				break;
			}

			// dispatch new task
			if(brute_newtask(disp) == ERROR) break;
		}
	}

	// end of passwords, wait all tasks
	while((task = brute_wait_task(disp))) dispatcher_free_slot(disp,task->pid);

	// show cracked
	session_show_status(session);

	// cleanup linked list
	dispatcher_free(disp);

	if(session->abort_bruteforce != ABORTED) return SUCCESS;
	else return ERROR;
}

/*
** Private methods
*/

s_task *brute_wait_task(s_dispatcher *disp)
{
	s_task		*task = dispatcher_waitpids(disp);

	// check the result
	if(!task) return NULL;

	if(task->retcode == VALID_LOGIN)
		session_new_valid_login(session,((s_brute_param *)task->argument)->user,((s_brute_param *)task->argument)->passwd);
	else if(task->retcode == BRUTE_ERROR)
		OUTPUT_ERROR("Error detected: %s/%s\n",((s_brute_param *)task->argument)->user,((s_brute_param *)task->argument)->passwd);

	// free argument
	free(((s_brute_param *)task->argument)->user);
	free(((s_brute_param *)task->argument)->passwd);
	free(task->argument);

	session_task_finished(session);

	return task;
}

int brute_newtask(s_dispatcher *disp)
{
	s_brute_param	*task_param = (s_brute_param *)do_malloc(sizeof(s_brute_param));

	// fill params
	task_param->libnet = session->libnet;
	task_param->params = session->module_params;
	task_param->module = session->module;

	// if no username/password, EOF
	if(!session_next_pair(session)) return ERROR;

	task_param->user = do_strdup(session_current_user(session));
	task_param->passwd = do_strdup(session_current_pass(session));

	// dispatch new task
	return dispatcher_newtask(disp,brute_task,(void *)task_param);
}

// SIGQUIT handler, used to abort the brute force
void brute_abort_request(int s) 
{ 
	session->abort_bruteforce = ABORT_REQUEST; 

	OUTPUT_ERROR("aborting...\n");
	
	// ignore further abort and status requests
	signal(SIGINT,SIG_IGN);
	signal(SIGQUIT,SIG_IGN);
}

// SIGINT handler, used to display status
void brute_status_request(int s)
{
	session_show_status(session);

	// re-set signal handler
	signal(SIGINT,brute_status_request);
}

/* this function controls the module, in the fork()'d proccess */
int brute_task(void *param)
{
	s_brute_param	*task_param = (s_brute_param *)param;
	s_mod_context	ctx;
	int		ret_code;

	// ignore SIGINT and SIGQUIT
	signal(SIGINT,SIG_IGN);
	signal(SIGQUIT,SIG_IGN);

	// setup module context
	ctx.libnet = task_param->libnet;
	ctx.params = task_param->params;
	ctx.user = task_param->user;
	ctx.passwd = task_param->passwd;

	// try to connect to the target	
	net_connect(task_param->libnet);

	// execute module action
	ret_code = task_param->module->module_function(&ctx);
	net_close(ctx.libnet);
	net_free(ctx.libnet);

	return ret_code;
}
