/*
 * Huk: a multi-threaded modular bruteforce tool 
 *
 * Written by Carlos Barros <barros [at] barrossecurity [dot] com>
 *
 * Copyright (C) 2006, Barrossecurity
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>

#include <libhuk/common.h>
#include <libhuk/wrappers.h>
#include <libhuk/debug.h>

#include "dispatcher.h"

/*=[ Prototypes ]=*/
void dispatcher_timed_out(int);

/*
** Public methods
*/

s_dispatcher *dispatcher_new(int maxtasks, int task_timeout)
{
	s_dispatcher	*newfork = (s_dispatcher *)do_malloc(sizeof(s_dispatcher));
	
	newfork->tasks = (s_task *)do_calloc(maxtasks,sizeof(s_task));
	newfork->maxtasks = maxtasks;
	newfork->task_timeout = task_timeout;

	return newfork;
}

void dispatcher_free(s_dispatcher *dispatcher)
{
	free(dispatcher->tasks);
	free(dispatcher);
}

int dispatcher_slotavailable(s_dispatcher *dispatcher)
{
	int	i;

	for(i=0;i<dispatcher->maxtasks;i++)
		if(dispatcher->tasks[i].pid == 0) 
		{
			debug(3,"dispatcher_slotavailable: slot %d availabe",i);
			return i;
		}

	return -1;
}

int dispatcher_newtask(s_dispatcher *dispatcher,int (*task)(void *),void *args)
{
	int	pid;
	int	slot = dispatcher_slotavailable(dispatcher);

	if(slot == -1) return ERROR;

	switch((pid = fork()))
	{
		case -1: abort("fork: %s",strerror(errno));
		break;
		case 0:
			// set timeout
			signal(SIGALRM,dispatcher_timed_out);
			alarm(dispatcher->task_timeout);

			// Dispatch task
			exit(task(args));
		break;
		default:
			debug(3,"dispatcher_newtask: new task (%d) created",pid);

			dispatcher->tasks[slot].pid = pid;
			dispatcher->tasks[slot].argument = args;
			
			return SUCCESS;
		break;
	}
}

s_task *dispatcher_waitpids(s_dispatcher *dispatcher)
{
	int			status,pid,i;

	// check if theres some child process pending
	for(i=0;i<dispatcher->maxtasks;i++)
		if(dispatcher->tasks[i].pid) break;
	if(i == dispatcher->maxtasks) return NULL;
	
	// Wait for one child to exit
	pid = wait(&status);
	
	// locate the exited pid
	for(i=0;i<dispatcher->maxtasks;i++) 
		if(dispatcher->tasks[i].pid == pid) 
		{
			dispatcher->tasks[i].retcode = WEXITSTATUS(status);
			debug(3,"dispatcher_waitpids: task %d exited with code %d",pid,dispatcher->tasks[i].retcode);
			break;
		}
	return &dispatcher->tasks[i];
}

void dispatcher_free_slot(s_dispatcher *dispatcher, int pid)
{
	int	i;

	// locate slot
	for(i=0;i<dispatcher->maxtasks;i++)
		if(dispatcher->tasks[i].pid == pid)
		{
			bzero(&dispatcher->tasks[i],sizeof(s_task));
			debug(3,"dispatcher_free_slot: slot %d free'd",i);
			break;
		}
}

/*
** Private methods
*/

void dispatcher_timed_out(int s) { exit(ERROR); }
