/*
 * Huk: a multi-threaded modular bruteforce tool 
 *
 * Written by Carlos Barros <barros [at] barrossecurity [dot] com>
 *
 * Copyright (C) 2006, Barrossecurity
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/select.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "common.h"
#include "socks4.h"
#include "socks5.h"
#include "wrappers.h"
#include "libnet.h"
#include "debug.h"

/*=[ Prototypes ]=*/
s_libnet_hp *net_resolv(char *);

/*
** Public methods
*/

s_libnet *net_new(char *host, int port)
{
	char		*proxy_env = NULL;

	s_libnet	*n_libnet = do_calloc(sizeof(s_libnet),sizeof(char));

	n_libnet->hp = net_resolv(host);
	n_libnet->port = port;
	n_libnet->type = DIRECT;

	// check for proxy requests
	if((proxy_env = getenv("SOCKS5_PROXY"))) n_libnet->type = SOCKS5;
	else if((proxy_env = getenv("SOCKS4_PROXY"))) n_libnet->type = SOCKS4;

	if(proxy_env)
	{
		char	*suser		= NULL,
			*spass		= NULL,
			*shost		= NULL,
			*sport		= NULL;

		// parse proxy env
		// first, look at @ sign
		shost = strchr(proxy_env,'@');
		if(shost)
		{
			// parse auth information
			*shost++ = 0x00;
			if((spass = strchr(proxy_env,':')))
			{
				*spass++ = 0x00;
				if(*spass && (*(suser = proxy_env)));
				else { spass = NULL; suser = NULL; }
			}
			if(!spass) warning("net_new: invalid socks auth information, ignoring auth ...")
			else if(n_libnet->type == SOCKS4) warning("net_new: SOCKS4 does not support authentication, ignoring auth ...");
		}
		else shost = proxy_env;

		// split host:port
		if(!(sport = strchr(shost,':'))) 
		{
			warning("net_new: socks proxy port not specified, ignoring socks ...");
			n_libnet->type = DIRECT;
		}
		else
		{
			*sport++ = 0;

			if(!*shost)
			{
				warning("net_new: invalid socks proxy host, ignoring socks ...");
				n_libnet->type = DIRECT;
			}

			n_libnet->shp = net_resolv(shost);
			n_libnet->sport = atoi(sport);
			n_libnet->suser = suser;
			n_libnet->spass = spass;

			// validate port
			if(n_libnet->sport <= 0)
			{
				warning("net_new: invalid socks proxy port, ignoring socks ...");
				n_libnet->type = DIRECT;
			}
			else 
			{
				OUTPUT("Using SOCKS%d proxy @ %s:%d",
					(n_libnet->type == SOCKS5)?5:4,n_libnet->shp->h_name,n_libnet->sport);
				if((n_libnet->type == SOCKS5) && (n_libnet->spass))
					OUTPUT(" authenticated by %s:%s\n",n_libnet->suser,n_libnet->spass)
				else OUTPUT("\n");
			}
		}
	}

	return n_libnet;
}

void net_free(s_libnet *slib) 
{ 
	if(slib->hp) { free(slib->hp->h_name); free(slib->hp); }
	if(slib->shp) { free(slib->shp->h_name); free(slib->shp); }
	free(slib); 
}

void net_close(s_libnet *slib) { close(slib->sock); }

void net_connect(s_libnet *slib)
{
	s_libnet_hp		*hp;
	int			port;
	struct sockaddr_in 	server;
	int             	sock;

	if(slib->type == DIRECT) { hp = slib->hp; port = slib->port; }
	else { hp = slib->shp; port = slib->sport; }

	server.sin_family = AF_INET;
	server.sin_addr = hp->h_address;
	server.sin_port = htons(port);

	// try to connect
	sock = do_socket(PF_INET,SOCK_STREAM,0);
	do_connect(sock,(struct sockaddr *)&server, sizeof(server));
	debug(3,"net_connect: connected to %s:%d",hp->h_name,port);

	slib->sock = sock;

	if(slib->type == SOCKS4) socks4_connect(slib);
	else if(slib->type == SOCKS5) socks5_connect(slib);
}


int net_write(s_libnet *slib, void *buffer, int size) 
{ 
	int	total = 0;
	int	fd = slib->sock;

	while(total < size)
	{
		int parc = write(fd,(void *)(buffer+total),size-total);
		if(parc == -1) 
		{
			return ERROR;
		}
		total += parc;
	}
	return SUCCESS;
}

int net_read(s_libnet *slib, void **buffer, int max_len) 
{ 
	int		total,parc,retval;
	int		fd = slib->sock;
	char		tmpbuffer[256];
	void		*dptr,*data = NULL;
	fd_set		rfds;
	struct timeval	tv;


	for(total = 0;!max_len || (total < max_len);)
	{
		FD_ZERO(&rfds);
		FD_SET(fd,&rfds);
		// Set timeout
		tv.tv_sec = 0;
		tv.tv_usec = READ_TIMEOUT;

		retval = select(fd+1,&rfds,NULL,NULL,&tv);
		if(retval == -1) abort("select: %s",strerror(errno))
		else if(retval)
		{
			if(!FD_ISSET(fd,&rfds)) continue;	// just in case

			parc = read(fd,tmpbuffer,(max_len)?(((max_len - total)>=256)?256:max_len - total):256);
			switch(parc)
			{
				case -1:
					if(errno == EWOULDBLOCK || errno == EAGAIN) continue;

					if(data) free(data);
					return -1;
				break;
				case 0: 
					*buffer = data;
					return total;
				break;
				default:

					// Ajust alloc'd mem and pointers
					data = do_realloc(data,total+parc+1);
					dptr = data + total;
					total += parc;
					memcpy(dptr,tmpbuffer,parc);
				break;
			}
		}
		else break;
	}

	*buffer = data;
	return total;
}

/*
** Private methods
*/

s_libnet_hp *net_resolv(char *host)
{
	struct hostent	*hp;
	s_libnet_hp	*n_hp = (s_libnet_hp *)do_malloc(sizeof(s_libnet_hp));

	// resolve address
	hp = gethostbyname(host);
	if(!hp) abort("gethostbyname: %s (%s)",hstrerror(h_errno),host);

	// fill libnet_hp struct
	memcpy(&n_hp->h_address,hp->h_addr,4);
	n_hp->h_name = do_strdup(hp->h_name);

	return n_hp;
}
