/*
 * Huk: a multi-threaded modular bruteforce tool 
 *
 * Written by Carlos Barros <barros [at] barrossecurity [dot] com>
 *
 * Copyright (C) 2006, Barrossecurity
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>

#include "common.h"
#include "socks5.h"
#include "libnet.h"
#include "wrappers.h"
#include "debug.h"

/*=[ Prototypes ]=*/
void socks5_authenticate(s_libnet *);

/*
** Public methods
*/

void socks5_connect(s_libnet *slib)
{
	unsigned char		*buffer;
	int			read_count;
	s_socks5_handshake	s_handshake;
	s_socks5_request	s_request;

	// fill handshake request
	s_handshake.ver				= SOCKS5_VERSION;			// SOCKS version 5
	s_handshake.auth_count			= 1;					// Always 1
	if(slib->suser) s_handshake.auth_method	= SOCKS5_AUTH_USER_PASS;		// User/Password authentication
	else s_handshake.auth_method 		= SOCKS5_AUTH_NONE;			// No authentication

	// send initial handshake
	net_write(slib,(void *)&s_handshake,sizeof(s_socks5_handshake));

	debug(3,"socks5_connect: initial handshake sent, waiting for reply ...");

	// check response
	while(!(read_count = net_read(slib,(void **)&buffer,2)));
	if(read_count == -1) abort("socks5_connect: %s",strerror(errno));

	debug(3,"socks5_connect: response received - r[0]=%u , r[1]=%u",buffer[0],buffer[1]);

	if(buffer[0] == SOCKS5_VERSION)
	{
		if(buffer[1] == SOCKS5_RESPONSE_AUTH_INVALID) 
			abort("socks5_connect: authentication method not supported");
	}
	else abort("socks5_connect: unexpected reply");

	debug(3,"socks5_connect: request accepted");
	free(buffer);

	if(slib->suser) socks5_authenticate(slib);

	// fill request
	s_request.ver		= SOCKS5_VERSION;			// SOCKS version 5
	s_request.cmd		= SOCKS5_CMD_TCP_CONNECT;		// TCP connect
	s_request.res		= 0;					// Reserver. Must be 0
	s_request.addr_type	= SOCKS5_ADDR_IPV4;			// IPv4 addr
	s_request.ip		= slib->hp->h_address.s_addr;		// IP
	s_request.port		= htons(slib->port);			// Port

	// send request
	net_write(slib,(void *)&s_request,sizeof(s_socks5_request));

	debug(3,"socks5_connect: connection request sent, waiting for reply ...");

	// check response
	while(!(read_count = net_read(slib,(void **)&buffer,10)));
	if(read_count == -1) abort("socks5_connect: %s",strerror(errno));

	debug(3,"socks5_connect: response received - r[0]=%u , r[1]=%u",buffer[0],buffer[1]);

	if(buffer[0] == SOCKS5_VERSION)
	{
		switch(buffer[1])
		{
			case SOCKS5_RESPONSE_GRANTED: 

				debug(3,"socks5_connect: connection established");
				
				free(buffer); 
				return;
			break;
			case SOCKS5_RESPONSE_FAILURE: abort("socks5_connect: general failure");
			break;
			case SOCKS5_RESPONSE_NOT_ALLOWED: abort("socks5_connect: connection not allowed by ruleset");
			break;
			case SOCKS5_RESPONSE_NET_UNREACHABLE: abort("socks5_connect: network unreachable");
			break;
			case SOCKS5_RESPONSE_HOST_UNREACHABLE: abort("socks5_connect: host unreachable");
			break;
			case SOCKS5_RESPONSE_REFUSED: abort("socks5_connect: connection refused by destination host");
			break;
			case SOCKS5_RESPONSE_TTL_EXPIRED: abort("socks5_connect: TTL expired");
			break;
			case SOCKS5_RESPONSE_CMD_NOT_SUPPORTED: abort("socks5_connect: command not supported / protocol error");
			break;
			case SOCKS5_RESPONSE_ADDR_NOT_SUPPORTED: abort("socks5_connect: address type not supported");
			break;
		}
	}
	abort("socks5_connect: unexpected reply");
}

/*
** Private methods
*/

void socks5_authenticate(s_libnet *slib)
{
	unsigned char		*buffer;
	unsigned char		*auth_request;
	int			read_count;

	debug(3,"socks5_authenticate: authenticating ...");

	// aloc request buffer
	auth_request = (unsigned char *)do_malloc(strlen(slib->suser)+strlen(slib->spass)+3);

	// fill authentication request
	auth_request[0]	= 0x01;
	auth_request[1] = strlen(slib->suser);
	memcpy(&auth_request[2],slib->suser,strlen(slib->suser));
	auth_request[2+strlen(slib->suser)] = strlen(slib->spass);
	memcpy(&auth_request[3+strlen(slib->suser)],slib->spass,strlen(slib->spass));

	// send authentication request
	net_write(slib,(void *)auth_request,strlen(slib->suser)+strlen(slib->spass)+3);

	debug(3,"socks5_authenticate: authentication request sent, waiting for reply ... ");

	// check response
	while(!(read_count = net_read(slib,(void **)&buffer,2)));
	if(read_count == -1) abort("socks5_authenticate: %s",strerror(errno));

	debug(3,"socks5_authenticate: response received - r[0]=%u , r[1]=%u",buffer[0],buffer[1]);

	if(buffer[0] == SOCKS5_VERSION)
	{
		if(buffer[1]) abort("socks5_authenticate: authentication failed");
	}
	else abort("socks5_authenticate: unexpected reply");

	debug(3,"socks5_authenticate: authenticated");
	
	free(buffer);

}
