/*
 * Huk: a multi-threaded modular bruteforce tool 
 *
 * Written by Carlos Barros <barros [at] barrossecurity [dot] com>
 *
 * Copyright (C) 2006, Barrossecurity
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dlfcn.h>

#include <libhuk/common.h>
#include <libhuk/wrappers.h>
#include <libhuk/debug.h>

#include "config.h"
#include "modules.h"

s_linked_list		*modules;

/*
** Public methods
*/

void modules_get_modules(void)
{

	DIR			*m_dir;
	struct dirent		*m_file;
	struct stat		f_stat;
	char			filename[NAME_MAX+strlen(MODULES_DIRECTORY)+10]; // just to be safe =)
	s_module_entry		m_entry;

	// initialize modules list
	modules = list_new();

	m_dir = do_opendir(MODULES_DIRECTORY);
	while((m_file = readdir(m_dir)))
	{
		// do not worry, it wont overflow
		sprintf(filename,"%s%s",MODULES_DIRECTORY,m_file->d_name);

		// stat filename to check its type, as POSIX does not define
		// it in the struct dirent
		// PS: We will not use do_stat here because we do not want
		//     it to terminate the program on failure
		if(stat(filename,&f_stat) == -1) warning("modules_get_modules: unable to stat %s",filename)
		else
		{
			if(!S_ISREG(f_stat.st_mode)) continue;	// skip non regular files

			// check if it is a .so file (we check it by extension first)
			if(strcmp(&filename[strlen(filename)-3],".so")) 
			{
				warning("modules_get_modules: invalid module filename found (%s), ignoring ...",m_file->d_name);
				continue;
			}

			// now we try to dl_open it
			if(!(m_entry.dl_handle = dlopen(filename,RTLD_LAZY|RTLD_LOCAL)))
			{
				printf("%s\n",dlerror());
				warning("modules_get_modules: invalid(1) module found (%s), ignoring ...",m_file->d_name);
				continue;
			}

			// now, locate modules' definition structure
			if(!(m_entry.m_definition = (s_module_definition *)dlsym(m_entry.dl_handle,"module_definition")))
			{
				warning("modules_get_modules: invalid(2) module found (%s), ignoring ...",m_file->d_name);

				dlclose(m_entry.dl_handle);
				continue;
			}

			// got a valid module, append to the list
			list_append(&modules,(void *)&m_entry,sizeof(m_entry));
		}
	}

	(void)closedir(m_dir);
}

void modules_free(void) { list_clean(&modules,modules_free_module); }
void modules_free_module(void *m_entry) { dlclose(((s_module_entry *)m_entry)->dl_handle); }

void modules_list(void)
{
	s_linked_list		*mods;
	s_module_entry		*m_entry;

	mods = modules;
	while(mods)
	{
		m_entry = (s_module_entry *)mods->content;
		OUTPUT_ERROR("\t%s\t%s- %s\n",
			m_entry->m_definition->module_key,
			(strlen(m_entry->m_definition->module_key) < 8)?"\t":"",
			m_entry->m_definition->module_descr);
		mods = mods->next;
	}
}

void modules_usage(s_module_definition *module)
{
	OUTPUT_ERROR("Module: %s (%s)\n",module->module_descr,module->module_key);
	OUTPUT_ERROR("Params:");
	if(module->params_list)
	{
		int	i = 0;

		OUTPUT_ERROR("\n");
		while(module->params_list[i].param_name)
		{
			OUTPUT_ERROR("\t%s%s\t- %s%s\n",
				module->params_list[i].param_name,
				(strlen(module->params_list[i].param_name) < 8)?"\t":"",
				module->params_list[i].param_descr,
				(module->params_list[i].param_required)?" (required)":"");

			i++;
		}

	}
	else OUTPUT_ERROR(" No available params\n");

	exit(EXIT_FAILURE);
}

void modules_append_param(s_linked_list **list, char *param)
{
	s_module_params		m_param;
	char			*value;

	if((value = strchr(param,'='))) *value++ = 0;
	else return;
	
	m_param.param_name = do_strdup(param);
	m_param.param_value = do_strdup(value);

	debug(3,"modules_append_param: appending param %s = %s",param,value);	

	list_append(list,(void *)&m_param,sizeof(s_module_params));
}

void modules_clean_param(void *param)
{
	s_module_params		*m_param = (s_module_params *)param;

	if(m_param->param_name) free(m_param->param_name);
	if(m_param->param_value) free(m_param->param_value);
}

void modules_check_params(s_module_definition *module, s_linked_list **list)
{
	// search required params
	if(module->params_list)
	{
		int		i = 0;

		while(module->params_list[i].param_name)
		{
			if(module->params_list[i].param_required && !modules_get_param(list,module->params_list[i].param_name))
				abort("modules_check_params: required param (%s) not specified",module->params_list[i].param_name);
			i++;
		}
	}
}

s_module_definition *modules_locate(char *module)
{
	s_linked_list		*mods;
	s_module_entry		*m_entry;

	if(!module) return NULL;

	mods = modules;
	while(mods)
	{
		m_entry = (s_module_entry *)mods->content;

		if(!strcmp(m_entry->m_definition->module_key,module)) return m_entry->m_definition;
		mods = mods->next;
	}

	return NULL;
}

