/*
 * Huk: a multi-threaded modular bruteforce tool 
 *
 * Written by Carlos Barros <barros [at] barrossecurity [dot] com>
 *
 * Copyright (C) 2006, Barrossecurity
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <netdb.h>
#include <time.h>
#include <sys/mman.h>

#include <libhuk/common.h>
#include <libhuk/wrappers.h>
#include <libhuk/libnet.h>
#include <libhuk/list.h>
#include <libhuk/debug.h>

#include "dispatcher.h"
#include "wlist.h"
#include "session.h"
#include "modules.h"

/*=[ Prototypes ]=*/
void session_clean_logins(void *);
int session_calc_module_params_length(s_session *);

/*=[ Global vars ]=*/
extern int	debug_level;		// debug level, from debug.c

/*
** Public methods
*/

s_session *session_new(int ac, char **av)
{
	static s_session	session;		// session object. it is safe to use static here
	s_session_options	session_opts;
	int			opt;

	// define some default values
	bzero(&session_opts,sizeof(s_session_options));
	session_opts.passwd_list = DEFAULT_WORDLIST;
	session_opts.single_user = DEFAULT_USERNAME;

	bzero(&session,sizeof(s_session));
	session.name		= DEFAULT_SESSION_NAME;
	session.max_conn 	= DEFAULT_CONNECTIONS;
	session.conn_timeout	= DEFAULT_TIMEOUT;

	// create list of module params
	session.module_params = list_new();

	while((opt = getopt(ac,av,"m:w:W:u:U:C:h:p:c:t:x:vn:R")) != EOF)
	{
		switch(opt)
		{
			case 'n': 
					session.name = optarg;
					if(strlen(session.name) > MAX_SESSION_NAME) session.name[MAX_SESSION_NAME] = '\0';
			break;
			case 'R': session_opts.restore_flag = 1;
			break;
			case 'm': if(!(session.module = modules_locate(optarg))) session_abort("invalid module");
			break;
			case 'W': session_opts.single_passwd = NULL; session_opts.passwd_list = optarg;
			break;
			case 'w': session_opts.passwd_list = NULL; session_opts.single_passwd = optarg;
			break;
			case 'u': session_opts.user_list = NULL; session_opts.single_user = optarg;
			break;
			case 'U': session_opts.single_user = NULL; session_opts.user_list = optarg;
			break;
			case 'C': 
					session_opts.single_passwd = session_opts.passwd_list = session_opts.single_user = session_opts.user_list = NULL; 
					session_opts.combo_list = optarg;
			break;
			case 'h': session_opts.host = optarg;
			break;
			case 'p': if(!(session_opts.port = atoi(optarg))) session_abort("invalid port");
			break;
			case 'c': session.max_conn = atoi(optarg);
			break;
			case 't': session.conn_timeout = atoi(optarg);
			break;
			case 'x': 
					if(!session_opts.m_usage && strcasecmp(optarg,MODULE_USAGE_TRIGGER)) modules_append_param(&session.module_params,optarg);
					else session_opts.m_usage = 1;
			break;
			case 'v': if(debug_level < DEBUG_MAXLEVEL) debug_level++;
			break;
			default: session_abort(NULL);
		}
	}

	if(session_opts.restore_flag) session_restore(&session,&session_opts);
	else
	{
		// check params
		if(!session.module) session_abort("module not specified");

		// show module usage message?
		if(session_opts.m_usage) modules_usage(session.module);	// this function does not return
	}

	if(!session_opts.host) session_abort("host not specified");
	if((session.max_conn <= 0) || (session.max_conn > MAX_CONNECTIONS)) session_abort("invalid max connections");
	if(session.conn_timeout <= 0) session_abort("invalid connection timeout");

	// validate params
	modules_check_params(session.module,&session.module_params);
	if(session.module->module_validate_params && session.module->module_validate_params(&session.module_params) == ERROR) exit(EXIT_FAILURE);

	// create network interface
	if(!session_opts.port) session_opts.port = session.module->default_port;
	session.libnet = net_new(session_opts.host,session_opts.port);
	
	// open wordlist file
	session.wlist = wlist_open(session_opts.single_user,session_opts.single_passwd,
					session_opts.user_list,session_opts.passwd_list,session_opts.combo_list);

	if(session_opts.restore_flag) wlist_goto_entry(session.wlist,session.tried_logins);
	else
	{
		// create valid_logins list list
		session.valid_logins = list_new();
		session.tried_logins = 0;
	}

	// initialize timer
	session.start_time = time(NULL);

	return &session;
}

void session_free(s_session *session)
{
	list_clean(&session->valid_logins,session_clean_logins);
	net_free(session->libnet);
	list_clean(&session->module_params,modules_clean_param);
	wlist_close(session->wlist);
}

void session_save(s_session *session)
{
	s_session_file		file_header;
	s_linked_list		*x_list;
	int			curr_offset = sizeof(file_header);
	char			session_filename[MAX_SESSION_NAME+15];
	FILE			*fptr;

	// try to open session file
	sprintf(session_filename,"%s.huk_session",session->name);	// dont worry, wont overflow here! =)
	fptr = do_fopen(session_filename,"w");

	// fill file header
	file_header.port = session->libnet->port;
	file_header.max_conn = session->max_conn;
	file_header.conn_timeout = session->conn_timeout;
	file_header.tried_logins = session->tried_logins;
	file_header.elapsed_time = (time(NULL) - session->start_time) + session->elapsed_time;

	file_header.module_key = curr_offset; curr_offset += (strlen(session->module->module_key) + 1);
	file_header.host = curr_offset; curr_offset += (strlen(session->libnet->hp->h_name) + 1);

	if(session->wlist->user_list) { file_header.user_list = curr_offset; curr_offset += (strlen(session->wlist->user_list) + 1); }
	else file_header.user_list = 0;
	if(session->wlist->passwd_list) { file_header.passwd_list = curr_offset; curr_offset += (strlen(session->wlist->passwd_list) + 1); }
	else file_header.passwd_list = 0;
	if(session->wlist->combo_list) { file_header.combo_list = curr_offset; curr_offset += (strlen(session->wlist->combo_list) + 1); }
	else file_header.combo_list = 0;
	if(session->wlist->single_user) { file_header.single_user = curr_offset; curr_offset += (strlen(session->wlist->single_user) + 1); }
	else file_header.single_user = 0;
	if(session->wlist->single_passwd) { file_header.single_passwd = curr_offset; curr_offset += (strlen(session->wlist->single_passwd) + 1); }
	else file_header.single_passwd = 0;

	file_header.module_params_count = list_count(&(session->module_params));
	file_header.module_params = curr_offset; curr_offset += session_calc_module_params_length(session);
	file_header.valid_logins_count = list_count(&(session->valid_logins));
	file_header.valid_logins = curr_offset;

	// write header
	do_fwrite((void *)&file_header,sizeof(file_header),1,fptr);
	
	// write contents
	do_fwrite((void *)session->module->module_key,strlen(session->module->module_key) + 1,1,fptr);
	do_fwrite((void *)session->libnet->hp->h_name,strlen(session->libnet->hp->h_name) + 1,1,fptr);
	if(session->wlist->user_list) do_fwrite((void *)session->wlist->user_list,strlen(session->wlist->user_list) + 1,1,fptr);
	if(session->wlist->passwd_list) do_fwrite((void *)session->wlist->passwd_list,strlen(session->wlist->passwd_list) + 1,1,fptr);
	if(session->wlist->combo_list) do_fwrite((void *)session->wlist->combo_list,strlen(session->wlist->combo_list) + 1,1,fptr);
	if(session->wlist->single_user) do_fwrite((void *)session->wlist->single_user,strlen(session->wlist->single_user) + 1,1,fptr);
	if(session->wlist->single_passwd) do_fwrite((void *)session->wlist->single_passwd,strlen(session->wlist->single_passwd) + 1,1,fptr);

	// params
	x_list = session->module_params;
	while(x_list)
	{
		char	*data_aux;

		// param name
		data_aux = ((s_module_params *)x_list->content)->param_name;
		do_fwrite((void *)data_aux,strlen(data_aux),1,fptr);

		// =
		do_fwrite((void *)"=",1,1,fptr);

		// param value
		data_aux = ((s_module_params *)x_list->content)->param_value;
		do_fwrite((void *)data_aux,strlen(data_aux) + 1,1,fptr);

		x_list = x_list->next;
	}

	// valid logins
	x_list = session->valid_logins;
	while(x_list)
	{
		s_login_pair		*l_pair = (s_login_pair *)x_list->content;

		do_fwrite((void *)l_pair->username,strlen(l_pair->username) + 1,1,fptr);
		do_fwrite((void *)l_pair->password,strlen(l_pair->password) + 1,1,fptr);

		x_list = x_list->next;
	}

	fclose(fptr);
}

void session_restore(s_session *session, s_session_options *session_opts)
{
	char			session_filename[MAX_SESSION_NAME+15];
	char			*session_contents;
	s_session_file		*session_header;
	struct stat		f_stat;
	FILE			*fptr;
	int			i,offset_aux;

	// stat filename
	sprintf(session_filename,"%s.huk_session",session->name);	// dont worry, wont overflow here! =)
	do_stat(session_filename,&f_stat);
	
	// check if it is a regular file
	if(!S_ISREG(f_stat.st_mode)) abort("session_restore: %s is not a regular file",session_filename);
	// check for a minimum file size
	if(f_stat.st_size < sizeof(s_session_file)) abort("session_restore: invalid session file (too small)");

	// mmap session file into memory
	fptr = do_fopen(session_filename,"rb");
	session_contents = (char *)do_mmap(NULL,f_stat.st_size,PROT_READ,MAP_SHARED,fileno(fptr),0);
	fclose(fptr);

	// parse session file contents
	session_header = (s_session_file *)session_contents;

	session_opts->passwd_list = SESSION_FILE_GET_STR(passwd_list,0);
	session_opts->single_passwd = SESSION_FILE_GET_STR(single_passwd,0);
	session_opts->user_list = SESSION_FILE_GET_STR(user_list,0);
	session_opts->single_user = SESSION_FILE_GET_STR(single_user,0);
	session_opts->combo_list = SESSION_FILE_GET_STR(combo_list,0);
	session_opts->host = SESSION_FILE_GET_STR(host,0);
	session_opts->port = SESSION_FILE_GET_INT(port);

	if(!(session->module = modules_locate(SESSION_FILE_GET_STR(module_key,0)))) p_session_abort("invalid module");
	session->max_conn = SESSION_FILE_GET_INT(max_conn);
	session->conn_timeout = SESSION_FILE_GET_INT(conn_timeout);
	session->tried_logins = SESSION_FILE_GET_INT(tried_logins);
	session->elapsed_time = SESSION_FILE_GET_INT(elapsed_time);

	// read module params
	session->module_params = list_new();
	for(offset_aux=0,i=0;i<SESSION_FILE_GET_INT(module_params_count);i++)
	{
		char	*p_aux;

		p_aux = do_strdup(SESSION_FILE_GET_STR(module_params,offset_aux)); offset_aux += strlen(p_aux) + 1;
		modules_append_param(&(session->module_params),p_aux);
		free(p_aux);
	}

	// read valid logins list
	for(offset_aux=0,i=0;i<SESSION_FILE_GET_INT(valid_logins_count);i++)
	{
		char	*l_user, *l_pass;

		l_user = SESSION_FILE_GET_STR(valid_logins,offset_aux); offset_aux += strlen(l_user) + 1;
		l_pass = SESSION_FILE_GET_STR(valid_logins,offset_aux); offset_aux += strlen(l_pass) + 1;

		session_new_valid_login(session,l_user,l_pass);
	}

}

void session_new_valid_login(s_session *session, char *user, char *passwd)
{
	s_login_pair	l_pair;

	debug(3,"session_new_valid_login: adding new valid login (%s / %s)",user,passwd);

	l_pair.username = do_strdup(user);
	l_pair.password = do_strdup(passwd);
	list_append(&session->valid_logins,(void *)&l_pair,sizeof(s_login_pair));
}

void session_task_finished(s_session *session) { session->tried_logins++; }

int session_next_pair(s_session *session) { return wlist_read(session->wlist); }
char *session_current_user(s_session *session) { return session->wlist->username; }
char *session_current_pass(s_session *session) { return session->wlist->password; }

void session_show_status(s_session *session)
{
	int			valid_count;
	time_t			elapsed_time;
	
	elapsed_time = session->elapsed_time + (time(NULL) - session->start_time);
	valid_count = list_count(&session->valid_logins);

	OUTPUT("Discovered logins: %d\n",valid_count);
	if(valid_count)
	{
		s_linked_list		*aux = session->valid_logins;
		s_login_pair		*l_pair;

		while(aux)
		{ 
			l_pair = (s_login_pair *)aux->content;
			OUTPUT(" - %s/%s\n",l_pair->username,l_pair->password);
			aux = aux->next;
		}
	}

	OUTPUT("Tried: %d/%d (%.1f%%) in %u:%02u:%02u:%02u\n",
		session->tried_logins,
		session->wlist->entry_count,
		(((float)(float)session->tried_logins / (float)session->wlist->entry_count) * 100.0),
		(unsigned)(elapsed_time / 86400), 
		(unsigned int)(elapsed_time % 86400 / 3600), 
		(unsigned int)(elapsed_time % 3600 / 60), 
		(unsigned int)(elapsed_time % 60));

	if(session->wlist->username && session->wlist->password)
		OUTPUT("Last try: %s/%s\n",session->wlist->username,session->wlist->password);
}

/*
** Private methods
*/

void session_clean_logins(void *login)
{
	s_login_pair	*l_pair = (s_login_pair *)login;

	if(l_pair->username) free(l_pair->username);
	if(l_pair->password) free(l_pair->password);
}

int session_calc_module_params_length(s_session *session)
{
	s_linked_list		*x_list = session->module_params;
	s_module_params		*m_param;
	int			params_length = 0;

        while(x_list) 
	{ 
		m_param = (s_module_params *)x_list->content;

		params_length += strlen(m_param->param_name) + strlen(m_param->param_value) + 2; 
		x_list = x_list->next; 
	}
	return params_length;
}
