/*
 * Huk: a multi-threaded modular bruteforce tool 
 *
 * Written by Carlos Barros <barros [at] barrossecurity [dot] com>
 *
 * Copyright (C) 2006, Barrossecurity
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libhuk/common.h>
#include <libhuk/wrappers.h>

#include "wlist.h"

/*=[ Prototypes ]=*/
char *wlist_readline(FILE *, int, char *);

/*
** Public methods
*/

s_wlist *wlist_open(char *single_user, char* single_passwd, char *user_list, char *passwd_list, char *combo_list)
{ 
	s_wlist		*n_list;

	n_list = do_calloc(sizeof(s_wlist),sizeof(char));
	n_list->single_user = do_strdup(single_user);
	n_list->single_passwd = do_strdup(single_passwd);
	n_list->user_list = do_strdup(user_list);
	n_list->passwd_list = do_strdup(passwd_list);
	n_list->combo_list = do_strdup(combo_list);
	n_list->total_read = 0;

	if(combo_list) n_list->combo_fptr = do_fopen(combo_list,"r");
	else
	{
		if(user_list) n_list->user_fptr = do_fopen(user_list,"r");
		else if(single_user) snprintf(n_list->username,MAX_WORD,"%s",single_user);
		else abort("wlist_open: expected user(s) source");

		if(passwd_list) n_list->pass_fptr = do_fopen(passwd_list,"r");
		else if(single_passwd) snprintf(n_list->password,MAX_WORD,"%s",single_passwd);
		else abort("wlist_open: expected password(s) source");
	}
	
	// count number of entries
	n_list->entry_count = wlist_count_entries(n_list);
	if(!n_list->entry_count) abort("wlist_open: no username/password combinations found");

	return n_list;
}
void wlist_close(s_wlist *list) 
{
	// free alocated memory 
	if(list->single_user) free(list->single_user);
	if(list->single_passwd) free(list->single_passwd);
	if(list->user_list) free(list->user_list);
	if(list->passwd_list) free(list->passwd_list);
	if(list->combo_list) free(list->combo_list);

	// close file pointers
	if(list->user_fptr) fclose(list->user_fptr);
	if(list->pass_fptr) fclose(list->pass_fptr);
	if(list->combo_fptr) fclose(list->combo_fptr);

	free(list);
}

int wlist_read(s_wlist *list)
{
	if(list->combo_fptr)
	{
		char	*p;
		char	word[MAX_WORD];

		for(;;)
		{
			if(!wlist_readline(list->combo_fptr,MAX_WORD,word)) return 0;

			// split username and password
			if((p = strchr(word,':')))
			{
				*p++ = 0;
	
				// check combo
				if(word[0] && *p) break;
			}
			// invalid entry, ignore
		}

		snprintf(list->username,MAX_WORD,"%s",word);
		snprintf(list->password,MAX_WORD,"%s",p);
	}
	else
	{
		if(list->user_fptr && list->pass_fptr) // user list + password list
		{
			if(!list->username[0] && !list->password[0]) // inicial stage
			{
				if(!wlist_readline(list->user_fptr,MAX_WORD,list->username)) return 0;
				if(!wlist_readline(list->pass_fptr,MAX_WORD,list->password)) return 0;
			}
			else
			{
				if(!wlist_readline(list->pass_fptr,MAX_WORD,list->password))
				{
					do_fseek(list->pass_fptr,0,SEEK_SET);
					if(!wlist_readline(list->user_fptr,MAX_WORD,list->username)) return 0;
					if(!wlist_readline(list->pass_fptr,MAX_WORD,list->password)) return 0; // should never happen
				}
				
			}
		}
		else if(list->user_fptr) // user list + single password
			if(!wlist_readline(list->user_fptr,MAX_WORD,list->username)) return 0;
			else;
		else if(list->pass_fptr) // single user + password list
			if(!wlist_readline(list->pass_fptr,MAX_WORD,list->password)) return 0;
			else;
		else if(list->total_read) return 0; // single user + single password
	}
	
	return ++list->total_read;
}

void wlist_goto_entry(s_wlist *list, int entry_number)
{
	if(entry_number < 0) abort("wlist_goto_entry: invalid entry number");

	// reset list
	wlist_reset(list);
	if(entry_number) while(wlist_read(list) < entry_number);
}

void wlist_reset(s_wlist *list) 
{ 
	if(list->user_fptr) do_fseek(list->user_fptr,0,SEEK_SET); 
	if(list->pass_fptr) do_fseek(list->pass_fptr,0,SEEK_SET); 
	if(list->combo_fptr) do_fseek(list->combo_fptr,0,SEEK_SET); 

	if(list->user_fptr || list->combo_fptr) bzero(list->username,MAX_WORD);
	if(list->pass_fptr || list->combo_fptr) bzero(list->password,MAX_WORD);
	list->total_read = 0;
}

int wlist_count_entries(s_wlist *list)
{
	int	count = 0;

	while(wlist_read(list)) count++;
	wlist_reset(list);

	return count;
}

/*
** Private methods
*/

char *wlist_readline(FILE *fptr, int size, char *buffer)
{
	char	*p;

	if(!do_fgets(buffer,size,fptr)) return NULL;
	if((p = strchr(buffer,'\n'))) *p = '\0';

	return buffer;
}
